@@ -37,7 +37,7 @@
 
 static int check_csum_sblock(void *sb, int csum_size)
 {
-	char result[BTRFS_CSUM_SIZE];
+	u8 result[BTRFS_CSUM_SIZE];
 	u32 crc = ~(u32)0;
 
 	crc = btrfs_csum_data(NULL, (char *)sb + BTRFS_CSUM_SIZE,
@@ -100,20 +100,19 @@ static void print_sys_chunk_array(struct btrfs_super_block *sb)
 			if (cur_offset + len > array_size)
 				goto out_short_read;
 
-			print_chunk(buf, chunk);
 			num_stripes = btrfs_chunk_num_stripes(buf, chunk);
 			if (!num_stripes) {
-				printk(
-	    "ERROR: invalid number of stripes %u in sys_array at offset %u\n",
+				error(
+			"invalid number of stripes %u in sys_array at offset %u",
 					num_stripes, cur_offset);
 				break;
 			}
 			len = btrfs_chunk_item_size(num_stripes);
 			if (cur_offset + len > array_size)
 				goto out_short_read;
+			print_chunk(buf, chunk);
 		} else {
-			printk(
-		"ERROR: unexpected item type %u in sys_array at offset %u\n",
+			error("unexpected item type %u in sys_array at offset %u",
 				(u32)key.type, cur_offset);
 			break;
 		}
@@ -129,7 +128,7 @@ out:
 	return;
 
 out_short_read:
-	printk("ERROR: sys_array too short to read %u bytes at offset %u\n",
+	error("sys_array too short to read %u bytes at offset %u",
 			len, cur_offset);
 	free(buf);
 }
